import itertools
import logging
import re
import typing
from functools import lru_cache

from celus_nibbler.conditions import IsDateCondition
from celus_nibbler.coordinates import Coord, CoordRange, Direction, RelativeTo
from celus_nibbler.data_headers import DataHeaders
from celus_nibbler.errors import TableException
from celus_nibbler.parsers.non_counter.generic import BaseGenericArea
from celus_nibbler.sources import (
    AuthorsSource,
    DateSource,
    DimensionSource,
    ExtractParams,
    ItemIdSource,
    ItemSource,
    MetricSource,
    OrganizationSource,
    PublicationDateSource,
    TitleIdKind,
    TitleIdSource,
    TitleSource,
)

logger = logging.getLogger(__name__)


class CounterHeaderArea(BaseGenericArea):
    HEADER_DATE_COL_START = 1
    MAX_HEADER_ROW = 50
    TITLE_DOI_NAMES = {
        "Book DOI",
        "DOI",
        "Journal DOI",
    }
    TITLE_ISBN_NAMES = {"ISBN"}
    TITLE_ISSN_NAMES = {
        "ISSN",
        "Print ISSN",
        "Print_ISSN",
        "Printed_ISSN",
        "Printed ISSN",
    }
    TITLE_URI_NAMES = {
        "URI",
    }
    TITLE_EISSN_NAMES = {
        "Online ISSN",
        "Online_ISSN",
    }
    TITLE_PROPRIETARY_NAMES = {
        "Proprietary",
        "Proprietary ID",
        "Proprietary_ID",
        "Proprietary Identifier",
    }
    ITEM_DOI_NAMES: typing.Set[str] = set()
    ITEM_ISBN_NAMES: typing.Set[str] = set()
    ITEM_ISSN_NAMES: typing.Set[str] = set()
    ITEM_URI_NAMES: typing.Set[str] = set()
    ITEM_EISSN_NAMES: typing.Set[str] = set()
    ITEM_PROPRIETARY_NAMES: typing.Set[str] = set()
    ITEM_PUBLICATION_DATE_NAMES: typing.Set[str] = set()
    ITEM_AUTHORS_NAMES: typing.Set[str] = set()
    DIMENSION_NAMES_MAP = [
        ("Publisher", {"Publisher"}),
        ("Platform", {"Platform"}),
    ]

    TITLE_COLUMN_NAMES: typing.List[str] = []
    ITEM_COLUMN_NAMES: typing.List[str] = []
    ORGANIZATION_COLUMN_NAMES: typing.List[str] = []
    METRIC_COLUMN_NAMES: typing.List[str] = []

    ORGANIZATION_EXTRACT_PARAMS: ExtractParams = ExtractParams()
    METRIC_EXTRACT_PARAMS: ExtractParams = ExtractParams(
        # Totals are usually without metrics => skip it instead of failing
        on_validation_error=TableException.Action.SKIP,
    )
    DATE_EXTRACT_PARAMS: ExtractParams = ExtractParams()
    DATA_EXTRACT_PARAMS: ExtractParams = ExtractParams(
        default=0,
    )
    TITLE_EXTRACT_PARAMS: ExtractParams = ExtractParams()
    TITLE_IDS_EXTRACT_PARAMS: typing.Dict[str, ExtractParams] = {}
    ITEM_EXTRACT_PARAMS: ExtractParams = ExtractParams()
    ITEM_IDS_EXTRACT_PARAMS: typing.Dict[str, ExtractParams] = {}
    DIMENSIONS_EXTRACT_PARAMS: typing.Dict[str, ExtractParams] = {}

    @property
    @lru_cache
    def dimensions(self) -> typing.List[str]:
        return [e[0] for e in self.DIMENSION_NAMES_MAP]

    def date_check(self, coord):
        if re.match(r"^\d+$", coord.content(self.sheet)):
            return False
        return IsDateCondition(coord=coord).check(self.sheet)

    @property
    @lru_cache
    def header_row(self) -> CoordRange:
        """Find the line where counter header is"""
        # Right now it checks whether a single continuous date area is present

        for idx in range(self.MAX_HEADER_ROW):
            crange = CoordRange(
                Coord(idx, self.HEADER_DATE_COL_START, RelativeTo.START), Direction.RIGHT
            )

            matching = None
            twice = False
            last_content = None
            for cell in crange:
                try:
                    last_content = cell.content(self.sheet).strip() or last_content
                    if self.date_check(cell):
                        if matching is False:
                            twice = True
                        matching = True
                    else:
                        if matching is not None:
                            matching = False
                except TableException as e:
                    if e.action == TableException.Action.STOP:
                        break  # last cell reached
                    raise

            if matching is not None and not twice:
                return CoordRange(Coord(idx, 0, RelativeTo.START), Direction.RIGHT)

            if not matching and last_content == "Reporting_Period_Total":
                # When no months are found with the row which ends with
                # "Reporting_Period_Total" it means that
                # Exclude_Monthly_Details=True was used during report
                # creation -> we should raise a special exception in that case
                raise TableException(
                    row=idx,
                    sheet=self.sheet.sheet_idx,
                    reason="counter-header-without-monthly-details",
                )

        raise TableException(
            sheet=self.sheet.sheet_idx,
            reason="no-counter-header-found",
        )

    @property
    @lru_cache
    def data_headers(self):
        # First date which is parsed in the header
        for cell in itertools.islice(self.header_row, 1, None):
            if not self.date_check(coord=cell):
                continue

            first_data_cell = CoordRange(cell, Direction.DOWN)[1]
            return DataHeaders(
                roles=[
                    DateSource(
                        source=CoordRange(cell, Direction.RIGHT),
                        extract_params=self.DATE_EXTRACT_PARAMS,
                    ),
                ],
                data_cells=CoordRange(first_data_cell, Direction.RIGHT),
                data_direction=Direction.DOWN,
                data_extract_params=self.DATA_EXTRACT_PARAMS,
            )

    @property
    @lru_cache
    def title_source(self) -> typing.Optional[TitleSource]:
        if not self.TITLE_COLUMN_NAMES:
            # Name of title column not defined -> assumed the title is in first column
            return TitleSource(
                CoordRange(self.header_row[0], Direction.DOWN).skip(1),
                extract_params=self.TITLE_EXTRACT_PARAMS,
            )

        for cell in self.header_row:
            try:
                content = cell.content(self.sheet)
            except TableException as e:
                if e.action == TableException.Action.STOP:
                    break  # last cell reached

            if content.strip() in self.TITLE_COLUMN_NAMES:
                return TitleSource(
                    CoordRange(cell, Direction.DOWN).skip(1),
                    extract_params=self.TITLE_EXTRACT_PARAMS,
                )

        # title column was not found, assuming that first column is the title
        return TitleSource(
            CoordRange(self.header_row[0], Direction.DOWN).skip(1),
            extract_params=self.TITLE_EXTRACT_PARAMS,
        )

    @property
    @lru_cache
    def title_ids_sources(self) -> typing.Dict[str, TitleIdSource]:
        ids_sources = {}
        for cell in self.header_row:
            try:
                content = cell.content(self.sheet)
            except TableException as e:
                if e.action == TableException.Action.STOP:
                    break  # last cell reached

            # Validate name
            content = content.strip()

            if content in self.TITLE_DOI_NAMES:
                name = TitleIdKind.DOI
            elif content in self.TITLE_ISBN_NAMES:
                name = TitleIdKind.ISBN
            elif content.strip() in self.TITLE_ISSN_NAMES:
                name = TitleIdKind.Print_ISSN
            elif content.strip() in self.TITLE_EISSN_NAMES:
                name = TitleIdKind.Online_ISSN
            elif content.strip() in self.TITLE_PROPRIETARY_NAMES:
                name = TitleIdKind.Proprietary
            elif content.strip() in self.TITLE_URI_NAMES:
                name = TitleIdKind.URI
            else:
                # empty or other field
                continue

            source = TitleIdSource(
                name,
                CoordRange(cell, Direction.DOWN).skip(1),
                extract_params=self.TITLE_IDS_EXTRACT_PARAMS.get(name, ExtractParams()),
            )
            ids_sources[str(name)] = source

        return ids_sources

    @property
    @lru_cache
    def item_source(self) -> typing.Optional[ItemSource]:
        if not self.ITEM_COLUMN_NAMES:
            # Need to have item column defined
            return None

        for cell in self.header_row:
            try:
                content = cell.content(self.sheet)
            except TableException as e:
                if e.action == TableException.Action.STOP:
                    break  # last cell reached

            if content.strip() in self.ITEM_COLUMN_NAMES:
                return ItemSource(
                    CoordRange(cell, Direction.DOWN).skip(1),
                    extract_params=self.ITEM_EXTRACT_PARAMS,
                )

        # title column was not found, assuming that first column is the title
        return ItemSource(
            CoordRange(self.header_row[0], Direction.DOWN).skip(1),
            extract_params=self.ITEM_EXTRACT_PARAMS,
        )

    @property
    @lru_cache
    def item_ids_sources(self) -> typing.Dict[str, ItemIdSource]:
        if not self.ITEM_COLUMN_NAMES:
            # Need to have item column defined
            return {}

        ids_sources = {}
        for cell in self.header_row:
            try:
                content = cell.content(self.sheet)
            except TableException as e:
                if e.action == TableException.Action.STOP:
                    break  # last cell reached

            # Validate name
            content = content.strip()

            if content in self.ITEM_DOI_NAMES:
                name = TitleIdKind.DOI
            elif content in self.ITEM_ISBN_NAMES:
                name = TitleIdKind.ISBN
            elif content.strip() in self.ITEM_ISSN_NAMES:
                name = TitleIdKind.Print_ISSN
            elif content.strip() in self.ITEM_EISSN_NAMES:
                name = TitleIdKind.Online_ISSN
            elif content.strip() in self.ITEM_PROPRIETARY_NAMES:
                name = TitleIdKind.Proprietary
            elif content.strip() in self.ITEM_URI_NAMES:
                name = TitleIdKind.URI
            else:
                # empty or other field
                continue

            source = ItemIdSource(
                name,
                CoordRange(cell, Direction.DOWN).skip(1),
                extract_params=self.ITEM_IDS_EXTRACT_PARAMS.get(name, ExtractParams()),
            )
            ids_sources[str(name)] = source

        return ids_sources

    @property
    @lru_cache
    def item_authors_source(self) -> typing.Optional[AuthorsSource]:
        if not self.ITEM_AUTHORS_NAMES:
            return None

        for cell in self.header_row:
            try:
                content = cell.content(self.sheet)
            except TableException as e:
                if e.action == TableException.Action.STOP:
                    break  # last cell reached

            if content.strip() in self.ITEM_AUTHORS_NAMES:
                return AuthorsSource(CoordRange(cell, Direction.DOWN).skip(1))

    @property
    @lru_cache
    def item_publication_date_source(self) -> typing.Optional[PublicationDateSource]:
        if not self.ITEM_PUBLICATION_DATE_NAMES:
            return None

        for cell in self.header_row:
            try:
                content = cell.content(self.sheet)
            except TableException as e:
                if e.action == TableException.Action.STOP:
                    break  # last cell reached

            if content.strip() in self.ITEM_PUBLICATION_DATE_NAMES:
                return PublicationDateSource(CoordRange(cell, Direction.DOWN).skip(1))

    @property
    @lru_cache
    def dimensions_sources(self) -> typing.Dict[str, DimensionSource]:
        dim_sources = {}
        for cell in self.header_row:
            try:
                content = cell.content(self.sheet)
            except TableException as e:
                if e.action == TableException.Action.STOP:
                    break  # last cell reached

            content = content.strip()
            for dimension, names in self.DIMENSION_NAMES_MAP:
                if content.lower() in [e.lower() for e in names]:
                    dim_sources[dimension] = DimensionSource(
                        content,
                        CoordRange(Coord(cell.row + 1, cell.col), Direction.DOWN),
                        extract_params=self.DIMENSIONS_EXTRACT_PARAMS.get(
                            dimension, ExtractParams()
                        ),
                    )

        return dim_sources

    @property
    @lru_cache
    def organization_source(self) -> typing.Optional[OrganizationSource]:
        if not self.ORGANIZATION_COLUMN_NAMES:
            # Organization column not defined
            return None

        for cell in self.header_row:
            try:
                content = cell.content(self.sheet)
            except TableException as e:
                if e.action == TableException.Action.STOP:
                    break  # last cell reached

            if content.strip() in self.ORGANIZATION_COLUMN_NAMES:
                return OrganizationSource(
                    CoordRange(cell, Direction.DOWN).skip(1),
                    extract_params=self.ORGANIZATION_EXTRACT_PARAMS,
                )

        return None

    @property
    @lru_cache
    def metric_source(self) -> typing.Optional[MetricSource]:
        if not self.METRIC_COLUMN_NAMES:
            return None

        for cell in self.header_row:
            try:
                content = cell.content(self.sheet)
                if content and content.strip().lower() in [
                    e.lower() for e in self.METRIC_COLUMN_NAMES
                ]:
                    return MetricSource(
                        CoordRange(cell, Direction.DOWN).skip(1),
                        extract_params=self.METRIC_EXTRACT_PARAMS,
                    )
            except TableException as e:
                if e.action == TableException.Action.STOP:
                    raise TableException(
                        value=self.METRIC_COLUMN_NAMES,
                        row=cell.row,
                        sheet=self.sheet.sheet_idx,
                        reason="missing-metric-in-header",
                    )

        return None



import re

from celus_nibbler.conditions import RegexCondition
from celus_nibbler.coordinates import Coord
from celus_nibbler.data_headers import DataFormatDefinition

from . import c5


class Counter51NameMixin:
    @property
    def name(self):
        if self.data_format.name == "IR51_M1":
            code = "IR_M1"
        else:
            code = self.data_format.name[:-2]
        return f"static.counter51.{code}.Tabular"


class DR(Counter51NameMixin, c5.DR):
    data_format = DataFormatDefinition(name="DR51")
    heuristics = (
        RegexCondition(re.compile(r"^Report_Name$"), Coord(0, 0))
        & RegexCondition(re.compile(r"^Report_ID$"), Coord(1, 0))
        & RegexCondition(re.compile(r"^DR$"), Coord(1, 1))
        & RegexCondition(re.compile(r"^Release$"), Coord(2, 0))
        & RegexCondition(re.compile(r"^5.1$"), Coord(2, 1))
    )


class PR(Counter51NameMixin, c5.PR):
    data_format = DataFormatDefinition(name="PR51")
    heuristics = (
        RegexCondition(re.compile(r"^Report_Name$"), Coord(0, 0))
        & RegexCondition(re.compile(r"^Report_ID$"), Coord(1, 0))
        & RegexCondition(re.compile(r"^PR$"), Coord(1, 1))
        & RegexCondition(re.compile(r"^Release$"), Coord(2, 0))
        & RegexCondition(re.compile(r"^5.1$"), Coord(2, 1))
    )


class TR(Counter51NameMixin, c5.TR):
    data_format = DataFormatDefinition(name="TR51")
    heuristics = (
        RegexCondition(re.compile(r"^Report_Name$"), Coord(0, 0))
        & RegexCondition(re.compile(r"^Report_ID$"), Coord(1, 0))
        & RegexCondition(re.compile(r"^TR$"), Coord(1, 1))
        & RegexCondition(re.compile(r"^Release$"), Coord(2, 0))
        & RegexCondition(re.compile(r"^5.1$"), Coord(2, 1))
    )

    class Area(c5.TR.Area):
        DIMENSION_NAMES_MAP = [
            ("Access_Type", {"Access Type", "Access_Type"}),
            ("Access_Method", {"Access_Method", "Access Method"}),
            ("Data_Type", {"Data Type", "Data_Type"}),
            ("YOP", {"YOP", "Year of Publication", "Year_of_Publication"}),
            ("Publisher", {"Publisher"}),
            ("Platform", {"Platform"}),
        ]

    areas = [Area]


class IR(Counter51NameMixin, c5.IR):
    data_format = DataFormatDefinition(name="IR51")
    heuristics = (
        RegexCondition(re.compile(r"^Report_Name$"), Coord(0, 0))
        & RegexCondition(re.compile(r"^Report_ID$"), Coord(1, 0))
        & RegexCondition(re.compile(r"^IR$"), Coord(1, 1))
        & RegexCondition(re.compile(r"^Release$"), Coord(2, 0))
        & RegexCondition(re.compile(r"^5.1$"), Coord(2, 1))
    )


class IR_M1(Counter51NameMixin, c5.IR_M1):
    data_format = DataFormatDefinition(name="IR51_M1")
    heuristics = (
        RegexCondition(re.compile(r"^Report_Name$"), Coord(0, 0))
        & RegexCondition(re.compile(r"^Report_ID$"), Coord(1, 0))
        & RegexCondition(re.compile(r"^IR_M1$"), Coord(1, 1))
        & RegexCondition(re.compile(r"^Release$"), Coord(2, 0))
        & RegexCondition(re.compile(r"^5.1$"), Coord(2, 1))
    )

    class Area(c5.IR_M1.Area):
        DIMENSION_NAMES_MAP = [
            ("Publisher", {"Publisher"}),
            ("Platform", {"Platform"}),
            # C5.1 IR_M1 has Data_Type dimension which is not present in C5 IR_M1
            ("Data_Type", {"Data Type", "Data_Type"}),
        ]

    areas = [Area]



import re
import typing

from celus_nigiri import CounterRecord

from celus_nibbler.conditions import RegexCondition
from celus_nibbler.coordinates import Coord, RelativeTo
from celus_nibbler.data_headers import DataFormatDefinition
from celus_nibbler.errors import TableException
from celus_nibbler.parsers.base import BaseTabularParser
from celus_nibbler.validators import BaseValueModel

from . import CounterHeaderArea


class BaseCounter4Parser(BaseTabularParser):
    Area: typing.Type[CounterHeaderArea]

    # TODO perhaps implement some kind of YOP validator
    dimensions_validators: typing.Dict[str, typing.Type[BaseValueModel]] = {}

    HEADER_DATA_OFFSETS = (
        ("Created", 1),
        ("Reporting_Period", 3),
        ("Institution_Name", 6),
    )

    REPORT_TYPE_NAME: str

    @property
    def name(self):
        return f"static.counter4.{self.data_format.name}.Tabular"

    def get_extras(self) -> dict:
        area = self.Area(self.sheet, self.platform)

        # get header row
        try:
            col = area.header_row.coord.col
            row = area.header_row.coord.row
        except TableException:
            return {}

        res = {}
        # use offsets to extract data
        for key, offset in self.HEADER_DATA_OFFSETS:
            coord = Coord(row=row - offset, col=col, row_relative_to=RelativeTo.START)

            if coord.row < 0:
                # out of bounds
                continue
            try:
                res[key] = coord.content(self.sheet)
            except (TableException, IndexError):
                continue

        return res

    def analyze(self) -> typing.List[dict]:
        area = self.Area(self.sheet, self.platform)

        # get header row
        try:
            col = area.header_row.coord.col
            row = area.header_row.coord.row
        except TableException:
            return [{"code": "header-row-not-found"}]

        rtn_coord = Coord(col=col, row=row - 7, row_relative_to=RelativeTo.START)
        if rtn_coord.row < 0:
            return [{"code": "report-name-not-in-header"}]

        res = []
        rtn = rtn_coord.content(self.sheet)
        if self.REPORT_TYPE_NAME != rtn:
            res.append(
                {
                    "code": "wrong-report-type",
                    "found": rtn,
                    "expected": self.REPORT_TYPE_NAME,
                }
            )

        return res


class BR1(BaseCounter4Parser):
    REPORT_TYPE_NAME = "Book Report 1 (R4)"

    data_format = DataFormatDefinition(name="BR1")

    titles_to_skip: typing.List[str] = ["Total", "Total for all titles", ""]

    platforms = ["*"]
    heuristics = RegexCondition(re.compile(r"^Book Report 1\s*\(R4\)"), Coord(0, 0))

    class Area(CounterHeaderArea):
        TITLE_COLUMN_NAMES = ["Book"]

        def prepare_record(self, *args, **kwargs) -> CounterRecord:
            res = super().prepare_record(*args, **kwargs)
            res.metric = "Book Title Requests"
            return res

    areas = [Area]


class BR2(BaseCounter4Parser):
    REPORT_TYPE_NAME = "Book Report 2 (R4)"

    data_format = DataFormatDefinition(name="BR2")

    titles_to_skip: typing.List[str] = ["Total", "Total for all titles", ""]

    platforms = ["*"]
    heuristics = RegexCondition(re.compile(r"^Book Report 2\s*\(R4\)"), Coord(0, 0))

    class Area(CounterHeaderArea):
        TITLE_COLUMN_NAMES = ["Book"]

        def prepare_record(self, *args, **kwargs) -> CounterRecord:
            res = super().prepare_record(*args, **kwargs)
            res.metric = "Book Section Requests"
            return res

    areas = [Area]


class BR3(BaseCounter4Parser):
    REPORT_TYPE_NAME = "Book Report 3 (R4)"

    data_format = DataFormatDefinition(name="BR3")

    titles_to_skip: typing.List[str] = ["Total", "Total for all titles", ""]

    platforms = ["*"]
    heuristics = RegexCondition(re.compile(r"^Book Report 3\s*\(R4\)"), Coord(0, 0))

    class Area(CounterHeaderArea):
        TITLE_COLUMN_NAMES = ["Book"]
        METRIC_COLUMN_NAMES = ["Access Denied Category"]

    areas = [Area]


class DB1(BaseCounter4Parser):
    REPORT_TYPE_NAME = "Database Report 1 (R4)"

    data_format = DataFormatDefinition(name="DB1")

    titles_to_skip: typing.List[str] = ["Total", "Total for all databases", ""]
    platforms = ["*"]
    heuristics = RegexCondition(re.compile(r"^Database Report 1\s*\(R4\)"), Coord(0, 0))

    class Area(CounterHeaderArea):
        TITLE_COLUMN_NAMES = ["Database"]
        METRIC_COLUMN_NAMES = ["User Activity"]

    areas = [Area]


class DB2(BaseCounter4Parser):
    REPORT_TYPE_NAME = "Database Report 2 (R4)"

    data_format = DataFormatDefinition(name="DB2")

    titles_to_skip: typing.List[str] = ["Total", "Total for all databases", ""]
    platforms = ["*"]
    heuristics = RegexCondition(re.compile(r"^Database Report 2\s*\(R4\)"), Coord(0, 0))

    class Area(CounterHeaderArea):
        TITLE_COLUMN_NAMES = ["Database"]
        METRIC_COLUMN_NAMES = ["Access denied category"]

    areas = [Area]


class PR1(BaseCounter4Parser):
    REPORT_TYPE_NAME = "Platform Report 1 (R4)"

    data_format = DataFormatDefinition(name="PR1")

    platforms = ["*"]
    heuristics = RegexCondition(re.compile(r"^Platform Report 1\s*\(R4\)"), Coord(0, 0))

    class Area(CounterHeaderArea):
        METRIC_COLUMN_NAMES = ["User Activity"]

        @property
        def title_source(self):
            return None

    areas = [Area]


class JR1(BaseCounter4Parser):
    REPORT_TYPE_NAME = "Journal Report 1 (R4)"

    data_format = DataFormatDefinition(name="JR1")

    titles_to_skip: typing.List[str] = ["Total", "Total for all journals", ""]
    platforms = ["*"]
    heuristics = RegexCondition(re.compile(r"^Journal Report 1\s*\(R4\)"), Coord(0, 0))

    class Area(CounterHeaderArea):
        TITLE_COLUMN_NAMES = ["Journal"]

        def prepare_record(self, *args, **kwargs) -> CounterRecord:
            res = super().prepare_record(*args, **kwargs)
            res.metric = "FT Article Requests"
            return res

    areas = [Area]


class JR1a(BaseCounter4Parser):
    REPORT_TYPE_NAME = "Journal Report 1a (R4)"

    data_format = DataFormatDefinition(name="JR1a")

    titles_to_skip: typing.List[str] = ["Total", "Total for all journals", ""]
    platforms = ["*"]
    heuristics = RegexCondition(re.compile(r"^Journal Report 1a\s*\(R4\)"), Coord(0, 0))

    class Area(CounterHeaderArea):
        TITLE_COLUMN_NAMES = ["Journal"]

        def prepare_record(self, *args, **kwargs) -> CounterRecord:
            res = super().prepare_record(*args, **kwargs)
            res.metric = "Archive Article Requests"
            return res

    areas = [Area]


class JR1GOA(BaseCounter4Parser):
    REPORT_TYPE_NAME = "Journal Report 1 GOA (R4)"

    data_format = DataFormatDefinition(name="JR1GOA")

    titles_to_skip: typing.List[str] = ["Total", "Total for all journals", ""]
    platforms = ["*"]
    heuristics = RegexCondition(re.compile(r"^Journal Report 1 GOA\s*\(R4\)"), Coord(0, 0))

    class Area(CounterHeaderArea):
        TITLE_COLUMN_NAMES = ["Journal"]

        def prepare_record(self, *args, **kwargs) -> CounterRecord:
            res = super().prepare_record(*args, **kwargs)
            res.metric = "Gold Open Access Article Requests"
            return res

    areas = [Area]


class JR2(BaseCounter4Parser):
    REPORT_TYPE_NAME = "Journal Report 2 (R4)"

    data_format = DataFormatDefinition(name="JR2")

    titles_to_skip: typing.List[str] = ["Total", "Total for all journals", ""]
    heuristics = RegexCondition(re.compile(r"^Journal Report 2\s*\(R4\)"), Coord(0, 0))
    platforms = ["*"]

    class Area(CounterHeaderArea):
        TITLE_COLUMN_NAMES = ["Journal"]
        METRIC_COLUMN_NAMES = ["Access Denied Category"]

    areas = [Area]


class MR1(BaseCounter4Parser):
    REPORT_TYPE_NAME = "Multimedia Report 1 (R4)"

    data_format = DataFormatDefinition(name="MR1")

    titles_to_skip: typing.List[str] = ["Total", "Total for all collections", ""]
    platforms = ["*"]
    heuristics = RegexCondition(re.compile(r"^Multimedia Report 1\s*\(R4\)"), Coord(0, 0))

    class Area(CounterHeaderArea):
        DIMENSION_NAMES_MAP = [
            ("Content Provider", {"Content Provider"}),
            ("Platform", {"Platform"}),
        ]
        TITLE_COLUMN_NAMES = ["Collection"]

        def prepare_record(self, *args, **kwargs) -> CounterRecord:
            res = super().prepare_record(*args, **kwargs)
            res.metric = "Multimedia Full Content Unit Requests"
            return res

    areas = [Area]



import contextlib
import cProfile
import functools
import typing
from datetime import date, timedelta

from pydantic import ConfigDict, TypeAdapter

COMMON_DATE_FORMATS = [
    "%Y-%m",
    "%Y-%m-%d",
    "%b-%Y",
    "%Y-%b",
    "%b-%y",
    "%y-%b",
]


@contextlib.contextmanager
def profile(*args, **kwargs):
    profile = cProfile.Profile(*args, **kwargs)
    profile.enable()
    yield
    profile.disable()

    print()
    profile.print_stats()


PydanticConfig = ConfigDict(
    extra="forbid",
    frozen=False,
    undefined_types_warning=False,
    defer_build=True,
)


class JsonEncorder:
    @classmethod
    @functools.lru_cache
    def adapter(cls):
        return TypeAdapter(cls)

    def dict(self):
        return self.adapter().dump_python(self)

    def json(self):
        return self.adapter().dump_json(self).decode()

    def model_dump(self):
        """To become more compatible with basic BaseModel"""
        return self.dict()

    @classmethod
    def parse(cls, obj: typing.Dict[str, typing.Any]):
        return cls.adapter().validate_python(obj)


def start_month(in_date: date) -> date:
    return in_date.replace(day=1)


def end_month(in_date: date) -> date:
    if in_date.month == 12:
        return date(year=in_date.year + 1, month=1, day=1) - timedelta(days=1)
    else:
        return date(year=in_date.year, month=in_date.month + 1, day=1) - timedelta(days=1)


def colnum_to_colletters(colnum: int) -> str:
    colletters = ""
    while colnum > 0:
        colnum, remainder = divmod(colnum - 1, 26)
        colletters = chr(65 + remainder) + colletters
    return colletters



import typing
from enum import Enum

from celus_nigiri import CounterRecord

from celus_nibbler.utils import colnum_to_colletters

# TODO more verbose exception (e.g. write why does it failed)


class NibblerError(Exception):
    def dict(self) -> dict:
        return {"name": f"{self.__class__.__name__}"}


class RecordError(NibblerError):
    def __init__(self, idx: int, record: CounterRecord):
        self.record = record
        self.idx = idx

    def dict(self) -> dict:
        return {
            "idx": self.idx,
            "name": f"{self.__class__.__name__}",
            "start": self.record.start,
            "end": self.record.end,
            "metric": self.record.metric,
            "title": self.record.title,
            "organization": self.record.organization,
            "dimensions": self.record.dimension_data,
            "title_ids": self.record.title_ids,
        }

    def __eq__(self, other):
        if not isinstance(self, type(other)):
            return False
        return self.dict() == other.dict()


class WrongFileFormatError(NibblerError):
    def __init__(self, file, file_suffix):
        super().__init__()
        self.file = file
        self.file_suffix = file_suffix

    def __str__(self):
        return f'file: "{self.file}" with extension "{self.file_suffix}" is not supported.'

    def __repr__(self):
        return f"{self.__class__.__name__}(file={self.file},file_suffix={self.file_suffix})"

    def __eq__(self, other):
        if not isinstance(self, type(other)):
            return False
        return all(
            getattr(self, attr, None) == getattr(other, attr, None)
            for attr in ["file", "file_suffix"]
        )


class XlsError(NibblerError):
    def __init__(self, xls_exception):
        self.xls_exception = xls_exception

    def __str__(self):
        return f'XLS Exception: "{self.xls_exception}"'


class JsonException(NibblerError):
    """
    Error while parsing JSON format
    """

    def __init__(self, reason: str):
        self.reason = reason
        pass

    def __str__(self):
        return f"json format error: {self.reason}"

    def __repr__(self):
        return f"{self.__class__.__name__}(reason={self.reason})"


class TableException(NibblerError):
    """
    General exception informing about position in which the exception occured
    while parsing tabular format
    """

    class Action(str, Enum):
        FAIL = "fail"
        SKIP = "skip"
        STOP = "stop"
        PASS = "pass"

    ATTRS = ["value", "row", "col", "sheet", "reason"]

    def __init__(
        self,
        value=None,
        row: typing.Optional[int] = None,
        col: typing.Optional[int] = None,
        sheet: typing.Optional[int] = None,
        reason: str = "unspecified",
        action: Action = Action.FAIL,
    ):
        super().__init__()
        self.value = value
        self.row = row
        self.col = col
        self.sheet = sheet
        self.reason = reason
        self.action = action

    def __str__(self):
        laymancount_sheet = self.sheet + 1 if self.sheet is not None else "unspecified"
        laymancount_row = self.row + 1 if self.row is not None else "unspecified"
        laymancount_col = self.col + 1 if self.col is not None else "unspecified"
        colletters_explanation = (
            f' (col "{colnum_to_colletters(laymancount_col)}" if using software for table sheets)'
            if self.col is not None
            else ""
        )
        location = (
            f"sheet {laymancount_sheet}, row {laymancount_row}, col {laymancount_col}"
            f"{colletters_explanation}"
        )
        return f"""\
Problem with parsing your format has occured.
Value causing this exception: {self.value}
Location of this value: {location}.
Reason: {self.reason}.
"""

    def __repr__(self):
        return (
            f"{self.__class__.__name__}"
            f"(reason={self.reason},sheet={self.sheet},"
            f"row={self.row},col={self.col},value={self.value})"
        )

    def __eq__(self, other):
        if not isinstance(self, type(other)):
            return False
        return all(getattr(self, attr, None) == getattr(other, attr, None) for attr in self.ATTRS)

    def dict(self) -> dict:
        return {
            "name": f"{self.__class__.__name__}",
            "reason": self.reason,
            "sheet_idx": self.sheet,
            "row": self.row,
            "col": self.col,
            "value": self.value,
        }


class NoParserFound(NibblerError):
    def __init__(self, sheet_idx, *args):
        self.sheet_idx = sheet_idx

    def __repr__(self):
        return f"{self.__class__.__name__}(sheet_idx={self.sheet_idx})"

    def dict(self) -> dict:
        return {
            "name": f"{self.__class__.__name__}",
            "sheet_idx": self.sheet_idx,
        }


class NoParserForPlatformFound(NoParserFound):
    pass


class NoParserMatchesHeuristics(NoParserFound):
    def __init__(self, sheet_idx: int, parsers_info: dict, *args):
        self.sheet_idx = sheet_idx
        self.parsers_info = parsers_info

    def dict(self) -> dict:
        return {
            "name": f"{self.__class__.__name__}",
            "sheet_idx": self.sheet_idx,
            "parsers_info": self.parsers_info,
        }


class NoParserForFileTypeFound(NoParserFound):
    pass


class MultipleParsersFound(NibblerError):
    def __init__(self, sheet_idx, *args):
        self.sheet_idx = sheet_idx
        self.parsers = args

    def dict(self) -> dict:
        return {
            "name": f"{self.__class__.__name__}",
            "sheet_idx": self.sheet_idx,
            "parsers": list(self.parsers),
        }


class SameRecordsInOutput(RecordError):
    def __init__(self, clashing_idx: int, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.clashing_idx = clashing_idx

    def dict(self) -> dict:
        dct = super().dict()
        dct["clashing_idx"] = self.clashing_idx
        return dct


class NegativeValueInOutput(RecordError):
    pass


class MissingDateInOutput(RecordError):
    pass



import datetime
import itertools
import logging
import typing
from abc import ABCMeta, abstractmethod

from celus_nigiri import CounterRecord

from celus_nibbler import validators
from celus_nibbler.aggregator import BaseAggregator, NoAggregator
from celus_nibbler.conditions import BaseCondition
from celus_nibbler.data_headers import DataCells, DataFormatDefinition, DataHeaders
from celus_nibbler.errors import MissingDateInOutput, TableException
from celus_nibbler.reader import CsvSheetReader, JsonCounter5SheetReader, SheetReader
from celus_nibbler.sources import (
    AuthorsSource,
    DateSource,
    DimensionSource,
    ItemIdSource,
    ItemSource,
    MetricSource,
    OrganizationSource,
    PublicationDateSource,
    SpecialExtraction,
    TitleIdSource,
    TitleSource,
)
from celus_nibbler.utils import end_month, start_month

logger = logging.getLogger(__name__)


IDS = {
    "DOI",
    "ISBN",
    "Print_ISSN",
    "Online_ISSN",
    "Proprietary",
    "URI",
}


class BaseArea(metaclass=ABCMeta):
    aggregator: BaseAggregator = NoAggregator()

    def __init__(self, sheet: SheetReader, platform: str, **extras):
        self.sheet = sheet
        self.platform = platform
        self.extras = extras

    def prepare_record(
        self,
        record: CounterRecord,
    ) -> CounterRecord:
        return record

    def check_record(
        self,
        idx: int,
        record: CounterRecord,
    ):
        if record.start is None:
            raise MissingDateInOutput(idx, record)

    @abstractmethod
    def get_months(self) -> typing.List[datetime.date]:
        pass

    @property
    @abstractmethod
    def dimensions(self) -> typing.List[str]:
        pass

    @classmethod
    def make_areas(cls, sheet, platform: str, **extras) -> typing.List["BaseArea"]:
        return [cls(sheet, platform, **extras)]


class BaseJsonArea(BaseArea):
    pass


class BaseTabularArea(BaseArea):
    organization_source: typing.Optional[OrganizationSource] = None
    date_source: typing.Optional[DateSource] = None
    title_source: typing.Optional[TitleSource] = None
    title_ids_sources: typing.Dict[str, TitleIdSource] = {}
    item_source: typing.Optional[ItemSource] = None
    item_ids_sources: typing.Dict[str, ItemIdSource] = {}
    item_publication_date_source: typing.Optional[PublicationDateSource] = None
    item_authors_source: typing.Optional[AuthorsSource] = None
    dimensions_sources: typing.Dict[str, DimensionSource] = {}
    metric_source: typing.Optional[MetricSource] = None
    row_offset: int = 0
    max_areas_generated: typing.Optional[int] = 1
    min_valid_areas: int = 1

    def __init__(self, sheet: SheetReader, platform: str, initial_row_offset: int = 0, **extras):
        super().__init__(sheet, platform)
        # The offset of area is initially set to parser's offset
        self.row_offset = initial_row_offset

    @abstractmethod
    def find_data_cells(
        self,
        get_metric_name: typing.Callable[[str], str],
        check_metric_name: typing.Callable[[str, str], None],
    ) -> typing.List[DataCells]:
        pass

    @classmethod
    def make_areas(
        cls, sheet, platform: str, initial_row_offset: int = 0, **extras
    ) -> typing.List["BaseArea"]:
        logger.debug("Preparing areas for %s", cls)
        areas: typing.List["BaseTabularArea"] = []
        row_offset = initial_row_offset
        while not cls.max_areas_generated or len(areas) < cls.max_areas_generated:
            area = cls(sheet, platform, initial_row_offset=row_offset)
            try:
                # Try to detect the header
                area.find_data_cells(lambda x: x, lambda x, y: None)
            except TableException as e:
                # expect at least min_valid_areas => gracefully stop
                if e.reason == "no-header-data-found" and cls.min_valid_areas <= len(areas):
                    break
                # Otherwise raise and error
                raise

            logger.debug("Area with offset %d was prepared for %s", row_offset, cls)
            areas.append(area)

            # Recalculate offset
            row_offset = area.row_offset + 1

        return areas


class BaseHeaderArea(BaseTabularArea):
    @property
    @abstractmethod
    def data_headers(self) -> DataHeaders:
        pass

    def find_data_cells(
        self,
        get_metric_name: typing.Callable[[str], str],
        check_metric_name: typing.Callable[[str, str], None],
    ) -> typing.List[DataCells]:
        offset, data_cells = self.data_headers.detect_data_cells(
            self.sheet, self.row_offset, get_metric_name, check_metric_name
        )
        self.row_offset = offset  # Update detected offset
        return data_cells

    def _get_months_from_column(
        self,
        parser_row_offset: typing.Optional[int],
        area_row_offset: typing.Optional[int],
    ) -> typing.List[datetime.date]:
        """This will get all months from months column"""

        res = set()
        for idx in itertools.count(0):
            try:
                date = self.date_source.extract(
                    self.sheet,
                    idx,
                    parser_row_offset=parser_row_offset,
                    area_row_offset=area_row_offset,
                )
                res.add(date.replace(day=1))

            except TableException as e:
                if e.action == TableException.Action.SKIP:
                    continue
                elif e.action == TableException.Action.STOP:
                    break
                else:
                    raise

        return list(res)

    def _get_months_from_header(self) -> typing.List[datetime.date]:
        return [e.header_data.start for e in self.find_data_cells(lambda x: x, lambda x, y: None)]


class BaseDateArea(BaseHeaderArea):
    def get_months(self) -> typing.List[datetime.date]:
        return self._get_months_from_header()


class BaseParser(metaclass=ABCMeta):
    metrics_to_skip: typing.List[str] = ["Total"]
    available_metrics: typing.Optional[typing.List[str]] = None
    on_metric_check_failed: TableException.Action = TableException.Action.SKIP
    titles_to_skip: typing.List[str] = ["Total"]
    items_to_skip: typing.List[str] = ["Total"]
    dimensions_to_skip: typing.Dict[str, typing.List[str]] = {"Platform": ["Total"]}
    dimensions_validators: typing.Dict[str, typing.Type[validators.BaseValueModel]] = {}
    heuristics: typing.Optional[BaseCondition] = None
    metric_aliases: typing.Dict[str, str] = {}
    metric_value_extraction_overrides: typing.Dict[str, SpecialExtraction] = {}
    dimension_aliases: typing.Dict[str, str] = {}
    possible_row_offsets: typing.List[int] = [0]
    row_offset: int = 0

    @classmethod
    @abstractmethod
    def sheet_reader_classes(cls) -> typing.List[typing.Type[SheetReader]]:
        pass

    @property
    @abstractmethod
    def name(self) -> str:
        pass

    @property
    @abstractmethod
    def data_format(self) -> DataFormatDefinition:
        pass

    @property
    @abstractmethod
    def areas(self) -> typing.List[typing.Type[BaseArea]]:
        """Areas defined"""
        pass

    def get_areas(self) -> typing.List[BaseArea]:
        """List of all data source areas withing the sheet"""
        return list(
            itertools.chain(
                *(area_class.make_areas(self.sheet, self.platform) for area_class in self.areas)
            )
        )

    def __init__(self, sheet: SheetReader, platform: str):
        self.sheet = sheet
        self.platform = platform

    def heuristic_check(self) -> bool:
        if self.heuristics:
            for row_offset in self.possible_row_offsets:
                # Default Coord.row_relative_to is set to "area"
                # and we don't want to override coords in heuristic definition
                # to Coord.row_relative_to="parser", so we use area_row_offset
                # not parser_row_offset
                if self.heuristics.check(self.sheet, 0, row_offset):
                    # Set detect offset to be used later
                    self.row_offset = row_offset
                    return True
            return False
        return True

    @property
    @classmethod
    @abstractmethod
    def platforms(self) -> typing.List[str]:
        """List of available platforms (used for validation)"""
        return []

    @classmethod
    def check_platform(cls, platform: str):
        if "*" in cls.platforms:
            # '*' will match any platform
            return True
        if platform in cls.platforms:
            return True
        return False

    def _parse(self) -> typing.Generator[typing.Tuple[int, CounterRecord], None, None]:
        for idx, area in enumerate(self.get_areas()):

            def gen():
                for e in self.parse_area(area):
                    yield idx, e

            yield from gen()

    def get_metric_name(self, name: str) -> str:
        return self.metric_aliases.get(name, name)

    def get_dimension_name(self, name):
        return self.dimension_aliases.get(name, name)

    def parse(self) -> typing.Generator[typing.Tuple[int, CounterRecord], None, None]:
        for idx, record in self._parse():
            yield idx, record

    def parse_area(self, area) -> typing.Generator[typing.Tuple[int, CounterRecord], None, None]:
        return area.aggregator.aggregate(self._parse_area(area))

    @abstractmethod
    def _parse_area(
        self, area: BaseArea
    ) -> typing.Generator[typing.Tuple[int, CounterRecord], None, None]:
        pass

    def get_months(self) -> typing.List[typing.List[datetime.date]]:
        return [e.get_months() for e in self.get_areas()]

    def get_extras(self) -> dict:
        return {}

    def analyze(self) -> typing.List[dict]:
        return []


class BaseJsonParser(BaseParser):
    @classmethod
    def sheet_reader_classes(cls):
        return [JsonCounter5SheetReader]

    def parse(self) -> typing.Generator[typing.Tuple[int, CounterRecord], None, None]:
        for idx, record in super().parse():
            # process aliases
            record.metric = self.get_metric_name(record.metric) if record.metric else record.metric
            record.dimension_data = {
                self.get_dimension_name(key): value for key, value in record.dimension_data.items()
            }
            yield idx, record


class BaseTabularParser(BaseParser):
    def get_areas(self) -> typing.List[BaseArea]:
        # We need to override this method to inject row_offset
        return list(
            itertools.chain(
                *(
                    area_class.make_areas(
                        self.sheet, self.platform, initial_row_offset=self.row_offset
                    )
                    for area_class in self.areas
                )
            )
        )

    @classmethod
    def sheet_reader_classes(cls):
        return [CsvSheetReader]

    def _metric_check(
        self,
        metric,
        orig_metric,
        metrics_to_skip: typing.List[str],
        available_metrics: typing.Optional[typing.List[str]],
        on_metric_check_failed: TableException.Action,
    ):
        def error():
            raise TableException(
                sheet=self.sheet.sheet_idx,
                value=orig_metric,
                reason="wrong-metric-found",
                action=on_metric_check_failed,
            )

        # ignore case during skip
        if metric.lower() in metrics_to_skip:
            error()

        # keep case when processing available metrics
        if available_metrics and metric not in available_metrics:
            error()

    def _parse_area(
        self, area: BaseTabularArea
    ) -> typing.Generator[typing.Tuple[int, CounterRecord], None, None]:
        count = 0
        metrics_to_skip = [e.lower() for e in self.metrics_to_skip]
        try:

            def check_metric_name(value: str, orig_value: str):
                self._metric_check(
                    value,
                    orig_value,
                    metrics_to_skip,
                    self.available_metrics,
                    self.on_metric_check_failed,
                )

            data_cells = area.find_data_cells(self.get_metric_name, check_metric_name)
        except TableException as e:
            if e.action == TableException.Action.FAIL:
                raise

        # Area offset should be absolute
        area_row_offset = area.row_offset
        parser_row_offset = self.row_offset

        # Store title_ids and dimensions sources
        # so it can be reused in the for-cycle
        dimensions_sources = list(area.dimensions_sources.items())
        title_ids_sources = area.title_ids_sources
        item_ids_sources = area.item_ids_sources
        item_authors_source = area.item_authors_source
        item_publication_date_source = area.item_publication_date_source
        metrics_to_skip = [e.lower() for e in self.metrics_to_skip]
        titles_to_skip = [e.lower() for e in self.titles_to_skip]
        items_to_skip = [e.lower() for e in self.items_to_skip]
        dimensions_to_skip = {k: [e.lower() for e in v] for k, v in self.dimensions_to_skip.items()}
        metric_value_extraction_overrides = self.metric_value_extraction_overrides

        for idx in itertools.count(0):
            try:
                # iterates through ranges
                if area.title_source:
                    title = area.title_source.extract(
                        self.sheet,
                        idx,
                        parser_row_offset=parser_row_offset,
                        area_row_offset=area_row_offset,
                    )
                    if title is not None and title.lower() in titles_to_skip:
                        continue
                else:
                    title = None

                if area.item_source:
                    item = area.item_source.extract(
                        self.sheet,
                        idx,
                        parser_row_offset=parser_row_offset,
                        area_row_offset=area_row_offset,
                    )
                    if item is not None and item.lower() in items_to_skip:
                        continue
                else:
                    item = None

                if area.metric_source:
                    orig_metric = area.metric_source.extract(
                        self.sheet,
                        idx,
                        parser_row_offset=parser_row_offset,
                        area_row_offset=area_row_offset,
                    )
                    metric = self.get_metric_name(orig_metric)
                    self._metric_check(
                        metric,
                        orig_metric,
                        metrics_to_skip,
                        self.available_metrics,
                        self.on_metric_check_failed,
                    )
                else:
                    metric = None

                if area.organization_source:
                    organization = area.organization_source.extract(
                        self.sheet,
                        idx,
                        parser_row_offset=parser_row_offset,
                        area_row_offset=area_row_offset,
                    )
                else:
                    organization = None

                if area.date_source:
                    date = area.date_source.extract(
                        self.sheet,
                        idx,
                        parser_row_offset=parser_row_offset,
                        area_row_offset=area_row_offset,
                    )
                else:
                    date = None

                dimension_data = {}
                skip = False
                for k, dimension_source in dimensions_sources:
                    dimension_text = dimension_source.extract(
                        self.sheet,
                        idx,
                        validator=self.dimensions_validators.get(k),
                        parser_row_offset=parser_row_offset,
                        area_row_offset=area_row_offset,
                    )
                    if (
                        dimension_text is not None
                        and dimension_text.lower() in dimensions_to_skip.get(k, [])
                    ):
                        skip = True
                        break
                    dimension_data[self.get_dimension_name(k)] = dimension_text

                if skip:
                    continue

                title_ids = {}
                item_ids = {}
                for key in IDS:
                    if title_source := title_ids_sources.get(key):
                        value = title_source.extract(
                            self.sheet,
                            idx,
                            parser_row_offset=parser_row_offset,
                            area_row_offset=area_row_offset,
                        )
                        if value:
                            title_ids[title_source.last_key] = value

                    if item_source := item_ids_sources.get(key):
                        value = item_source.extract(
                            self.sheet,
                            idx,
                            parser_row_offset=parser_row_offset,
                            area_row_offset=area_row_offset,
                        )
                        if value:
                            item_ids[item_source.last_key] = value

                if item_publication_date_source:
                    item_publication_date = item_publication_date_source.extract(
                        self.sheet,
                        idx,
                        parser_row_offset=parser_row_offset,
                        area_row_offset=area_row_offset,
                    )
                else:
                    item_publication_date = None

                if item_authors_source:
                    item_authors = item_authors_source.extract(
                        self.sheet,
                        idx,
                        parser_row_offset=parser_row_offset,
                        area_row_offset=area_row_offset,
                    )
                else:
                    item_authors = None

            except TableException as e:
                if e.action == TableException.Action.SKIP:
                    continue
                if e.action == TableException.Action.STOP:
                    return
                else:
                    raise

            for data_cell in data_cells:
                try:
                    value_validator = metric_value_extraction_overrides.get(
                        data_cell.header_data.metric or metric or "",
                        SpecialExtraction.NO,
                    ).get_validator()

                    value = data_cell.value_source.extract(
                        self.sheet,
                        idx,
                        validator=value_validator,
                        parser_row_offset=parser_row_offset,
                        area_row_offset=area_row_offset,
                    )
                    record = CounterRecord(
                        value=round(value),
                        organization=organization,
                        metric=metric,
                        title=title,
                        item=item,
                        dimension_data=dimension_data,
                        title_ids=title_ids,
                        item_ids=item_ids,
                        item_publication_date=item_publication_date,
                        item_authors=item_authors,
                        start=start_month(date) if date else None,
                        end=end_month(date) if date else None,
                    )
                    record = data_cell.merge_into_record(record)
                    record = area.prepare_record(record)
                    logger.debug("Parsed %s", record)
                    area.check_record(count, record)
                    yield record
                    count += 1

                except TableException as e:
                    if e.action == TableException.Action.SKIP:
                        continue
                    if e.action == TableException.Action.STOP:
                        return
                    else:
                        raise



import typing
from dataclasses import field
from datetime import date
from enum import Enum

from pydantic import Field, ValidationError
from pydantic.dataclasses import dataclass, rebuild_dataclass
from typing_extensions import Annotated

from celus_nibbler import validators
from celus_nibbler.coordinates import Coord, CoordRange, SheetAttr, Value
from celus_nibbler.errors import TableException
from celus_nibbler.reader import SheetReader
from celus_nibbler.utils import JsonEncorder, PydanticConfig

Source = typing.Union[Coord, CoordRange, SheetAttr, Value]


class KindISBN(str, Enum):
    ISBN13 = "isbn13"
    ISBN10 = "isbn10"


class TitleIdKind(str, Enum):
    ISBN = "ISBN"
    Print_ISSN = "Print_ISSN"
    Online_ISSN = "Online_ISSN"
    Proprietary = "Proprietary"
    DOI = "DOI"
    URI = "URI"

    def __str__(self):
        return self.value

    def validator_class(
        self, opts: typing.Optional["IdValidatorOpts"] = None
    ) -> typing.Type[validators.BaseValueModel]:
        if self == TitleIdKind.ISBN:
            if not isinstance(opts, IdValidatorOptsISBN):
                opts = IdValidatorOptsISBN()
            return opts.validator_class()
        elif self == TitleIdKind.Print_ISSN:
            if not isinstance(opts, IdValidatorOptsISSN):
                opts = IdValidatorOptsISSN()
            return opts.validator_class()
        elif self == TitleIdKind.Online_ISSN:
            if not isinstance(opts, IdValidatorOptsEISSN):
                opts = IdValidatorOptsEISSN()
            return opts.validator_class()
        elif self == TitleIdKind.Proprietary:
            return validators.ProprietaryID
        elif self == TitleIdKind.DOI:
            return validators.DOI
        elif self == TitleIdKind.URI:
            return validators.URI
        else:
            raise NotImplementedError()


@dataclass(config=PydanticConfig)
class IdValidatorOptsISSN:
    type: typing.Literal[TitleIdKind.Print_ISSN] = TitleIdKind.Print_ISSN

    strict: bool = False

    def validator_class(self) -> typing.Type[validators.BaseValueModel]:
        return validators.StrictISSN if self.strict else validators.ISSN


@dataclass(config=PydanticConfig)
class IdValidatorOptsEISSN:
    type: typing.Literal[TitleIdKind.Online_ISSN] = TitleIdKind.Online_ISSN
    strict: bool = False

    def validator_class(self) -> typing.Type[validators.BaseValueModel]:
        return validators.StrictEISSN if self.strict else validators.EISSN


@dataclass(config=PydanticConfig)
class IdValidatorOptsISBN:
    type: typing.Literal[TitleIdKind.ISBN] = TitleIdKind.ISBN
    strict: bool = False
    isbn_kind: typing.Optional[KindISBN] = None

    def validator_class(self) -> typing.Type[validators.BaseValueModel]:
        if not self.strict:
            return validators.ISBN
        if not self.isbn_kind:
            return validators.StrictISBN
        elif self.isbn_kind == KindISBN.ISBN13:
            return validators.StrictISBN13
        elif self.isbn_kind == KindISBN.ISBN10:
            return validators.StrictISBN10

        raise NotImplementedError()


IdValidatorOpts = Annotated[
    typing.Union[
        IdValidatorOptsISBN,
        IdValidatorOptsISSN,
        IdValidatorOptsEISSN,
    ],
    Field(discriminator="type"),
]


class DateFormat(str, Enum):
    US = "us"  # US format 12/31/2022 (more common)
    EU = "eu"  # EU format 31/12/2022


class SpecialExtraction(str, Enum):
    NO = "no"
    COMMA_SEPARATED_NUMBER = "comma_separated_number"
    MINUTES_TO_SECONDS = "minutes_to_seconds"
    HOURS_TO_SECONDS = "hours_to_seconds"
    DURATION_TO_SECONDS = "duration_to_seconds"

    def get_validator(self) -> typing.Optional[typing.Type[validators.BaseValueModel]]:
        if self == SpecialExtraction.NO:
            return None
        elif self == SpecialExtraction.COMMA_SEPARATED_NUMBER:
            return validators.CommaSeparatedNumberValidator
        elif self == SpecialExtraction.MINUTES_TO_SECONDS:
            return validators.MinutesToSecondsValidator
        elif self == SpecialExtraction.HOURS_TO_SECONDS:
            return validators.HoursToSecondsValidator
        elif self == SpecialExtraction.DURATION_TO_SECONDS:
            return validators.DurationToSecondsValidator
        else:
            raise NotImplementedError()


class Role(str, Enum):
    VALUE = "value"
    DATE = "date"
    TITLE = "title"
    TITLE_ID = "title_id"
    DIMENSION = "dimension"
    METRIC = "metric"
    ORGANIZATION = "organization"
    VOID = "void"
    AUTHORS = "authors"
    PUBLICATION_DATE = "publication_date"


@dataclass(config=PydanticConfig)
class ExtractParams(JsonEncorder):
    regex: typing.Optional[typing.Pattern] = None
    default: typing.Optional[typing.Any] = None
    last_value_as_default: bool = False
    blank_values: typing.Tuple[typing.Any, ...] = field(default_factory=lambda: (None, ""))
    skip_validation: bool = False
    prefix: str = ""
    suffix: str = ""
    special_extraction: SpecialExtraction = SpecialExtraction.NO
    on_validation_error: TableException.Action = TableException.Action.FAIL
    max_idx: typing.Optional[int] = None


class ContentExtractorMixin:
    source: Source
    extract_params: ExtractParams
    cleanup_during_header_processing: bool
    fallback: typing.Optional["ContentExtractorMixin"]
    role: Role
    _last_sheet_idx = None
    _last_source = None
    _last_extracted = None
    _last_area_row_offset = None

    def content(
        self,
        sheet: SheetReader,
        source: Source,
        parser_row_offset: typing.Optional[int],
        area_row_offset: typing.Optional[int],
    ) -> typing.Any:
        content = source.content(sheet, parser_row_offset, area_row_offset)
        if regex := self.extract_params.regex:
            if extracted := regex.search(content):
                content = extracted.group(1)
            else:
                # Unable to extract data
                return None

        if prefix := self.extract_params.prefix:
            content = prefix + content

        if suffix := self.extract_params.suffix:
            content = content + suffix

        return content

    def extract(
        self,
        sheet: SheetReader,
        idx: int,
        validator: typing.Optional[typing.Type[validators.BaseValueModel]] = None,
        parser_row_offset: typing.Optional[int] = None,
        area_row_offset: typing.Optional[int] = None,
    ) -> typing.Any:
        try:
            value = self._extract(sheet, idx, validator, parser_row_offset, area_row_offset)
            if self.extract_params.last_value_as_default:
                self.extract_params.default = value

        except TableException as e:
            if e.action == TableException.Action.PASS and self.fallback:
                return self.fallback.extract(
                    sheet, idx, validator, parser_row_offset, area_row_offset
                )
            else:
                raise
        return value

    def get_validator(
        self, validator: typing.Optional[typing.Type[validators.BaseValueModel]]
    ) -> typing.Optional[typing.Type[validators.BaseValueModel]]:
        return validator or self.extract_params.special_extraction.get_validator() or self.validator

    def _extract(
        self,
        sheet: SheetReader,
        idx: int,
        validator: typing.Optional[typing.Type[validators.BaseValueModel]] = None,
        parser_row_offset: typing.Optional[int] = None,
        area_row_offset: typing.Optional[int] = None,
    ) -> typing.Any:
        source = self.source[idx]

        if self.extract_params.max_idx is not None:
            if idx > self.extract_params.max_idx:
                raise TableException(
                    row=getattr(source, "row_absolute", lambda x, y: None)(
                        parser_row_offset,
                        area_row_offset,
                    ),
                    col=getattr(source, "col", None),
                    sheet=sheet.sheet_idx,
                    reason="out-of-bounds",
                    action=TableException.Action.STOP,
                )

        if (
            source == self._last_source
            and self._last_sheet_idx == sheet.sheet_idx
            and self._last_area_row_offset == area_row_offset  # Should be in same area
            and self._last_extracted
        ):
            # Same value will be extracted from the same coord
            return self._last_extracted
        else:
            self._last_source = source

        try:
            content = self.content(sheet, source, parser_row_offset, area_row_offset)
            if validator := self.get_validator(validator):
                if self.extract_params.default is not None:
                    res = validators.gen_default_validator(
                        validator,
                        self.extract_params.default,
                        self.extract_params.blank_values,
                    )(value=content).value
                elif self.extract_params.skip_validation:
                    res = (content or "").strip()
                else:
                    res = validator(value=content).value

            else:
                res = content
        except ValidationError as e:
            if isinstance(self.source, Value):
                raise TableException(
                    value=self.source.value,
                    sheet=sheet.sheet_idx,
                    reason="wrong-value",
                    action=self.extract_params.on_validation_error,
                )
            elif isinstance(self.source, SheetAttr):
                raise TableException(
                    value=self.source.sheet_attr,
                    sheet=sheet.sheet_idx,
                    reason="wrong-sheet-attr",
                    action=self.extract_params.on_validation_error,
                )
            else:
                reason = e.title.lower()
                # Try to Extract reason from exception args if present
                if ctx := e.errors()[0].get("ctx"):
                    if error := ctx.get("error"):
                        if args := error.args:
                            reason = ",".join(args)
                raise TableException(
                    content,
                    row=getattr(source, "row_absolute", lambda x, y: None)(
                        parser_row_offset,
                        area_row_offset,
                    ),
                    col=getattr(source, "col", None),
                    sheet=sheet.sheet_idx,
                    reason=reason,
                    action=self.extract_params.on_validation_error,
                ) from e
        except IndexError as e:
            raise TableException(
                row=getattr(source, "row_absolute", lambda x, y: None)(
                    parser_row_offset,
                    area_row_offset,
                ),
                col=getattr(source, "col", None),
                sheet=sheet.sheet_idx,
                reason="out-of-bounds",
                action=TableException.Action.STOP,
            ) from e

        # update last extracted
        self._last_extracted = res
        self._last_sheet_idx = sheet.sheet_idx
        self._last_area_row_offset = area_row_offset
        return res

    @property
    def validator(self) -> typing.Optional[typing.Type[validators.BaseValueModel]]:
        return None


@dataclass(config=PydanticConfig)
class DimensionSource(JsonEncorder, ContentExtractorMixin):
    name: str
    source: Source
    extract_params: ExtractParams = field(default_factory=lambda: ExtractParams())
    fallback: typing.Optional["DimensionSource"] = None
    cleanup_during_header_processing: bool = True
    role: typing.Literal[Role.DIMENSION] = Role.DIMENSION

    @property
    def validator(self) -> typing.Optional[typing.Type[validators.BaseValueModel]]:
        return validators.Dimension


@dataclass(config=PydanticConfig)
class MetricSource(JsonEncorder, ContentExtractorMixin):
    source: Source
    extract_params: ExtractParams = field(default_factory=lambda: ExtractParams())
    fallback: typing.Optional["MetricSource"] = None
    cleanup_during_header_processing: bool = True
    role: typing.Literal[Role.METRIC] = Role.METRIC

    @property
    def validator(self) -> typing.Optional[typing.Type[validators.BaseValueModel]]:
        return validators.Metric


@dataclass(config=PydanticConfig)
class OrganizationSource(JsonEncorder, ContentExtractorMixin):
    source: Source
    extract_params: ExtractParams = field(default_factory=lambda: ExtractParams())
    fallback: typing.Optional["OrganizationSource"] = None
    cleanup_during_header_processing: bool = True
    role: typing.Literal[Role.ORGANIZATION] = Role.ORGANIZATION

    @property
    def validator(self) -> typing.Optional[typing.Type[validators.BaseValueModel]]:
        return validators.Organization


@dataclass(config=PydanticConfig)
class TitleSource(JsonEncorder, ContentExtractorMixin):
    source: Source
    extract_params: ExtractParams = field(default_factory=lambda: ExtractParams())
    fallback: typing.Optional["TitleSource"] = None
    cleanup_during_header_processing: bool = True
    role: typing.Literal[Role.TITLE] = Role.TITLE

    @property
    def validator(self) -> typing.Optional[typing.Type[validators.BaseValueModel]]:
        return validators.Title


@dataclass(config=PydanticConfig)
class TitleIdSource(JsonEncorder, ContentExtractorMixin):
    name: TitleIdKind
    source: Source
    extract_params: ExtractParams = field(default_factory=lambda: ExtractParams())
    cleanup_during_header_processing: bool = True
    fallback: typing.Optional["TitleIdSource"] = None
    validator_opts: typing.Optional[IdValidatorOpts] = None
    role: typing.Literal[Role.TITLE_ID] = Role.TITLE_ID

    def __post_init__(self):
        self._last_key = None

    @property
    def last_key(self) -> typing.Optional[str]:
        return self._last_key

    @property
    def validator(self) -> typing.Optional[typing.Type[validators.BaseValueModel]]:
        return self.name.validator_class(self.validator_opts)

    def extract(
        self,
        sheet: SheetReader,
        idx: int,
        validator: typing.Optional[typing.Type[validators.BaseValueModel]] = None,
        parser_row_offset: typing.Optional[int] = None,
        area_row_offset: typing.Optional[int] = None,
    ) -> typing.Any:
        self._last_key = None
        res = super().extract(sheet, idx, validator, parser_row_offset, area_row_offset)
        # Update last key from fallback
        if fallback := self.fallback:
            if last_key := fallback._last_key:
                self._last_key = last_key
                fallback._last_key = None  # clean fallback
                return res

        self._last_key = self.name
        return res


@dataclass(config=PydanticConfig)
class ItemSource(TitleSource):
    pass


@dataclass(config=PydanticConfig)
class ItemIdSource(TitleIdSource):
    pass


@dataclass(config=PydanticConfig)
class ComposedDate(JsonEncorder):
    year: "DateSource"
    month: "DateSource"


@dataclass(config=PydanticConfig)
class DateSource(JsonEncorder, ContentExtractorMixin):
    source: Source
    extract_params: ExtractParams = field(default_factory=lambda: ExtractParams())
    fallback: typing.Optional["DateSource"] = None
    cleanup_during_header_processing: bool = True
    preferred_date_format: DateFormat = DateFormat.US
    composed: typing.Optional[ComposedDate] = None
    date_pattern: typing.Optional[str] = None
    force_aligned: bool = False
    role: typing.Literal[Role.DATE] = Role.DATE

    @property
    def validator(self) -> typing.Optional[typing.Type[validators.BaseValueModel]]:
        return validators.Date

    def get_validator(
        self, validator: typing.Optional[typing.Type[validators.BaseValueModel]]
    ) -> typing.Optional[typing.Type[validators.BaseValueModel]]:
        if date_pattern := self.date_pattern:
            return validators.gen_date_format_validator(date_pattern)

        if self.preferred_date_format == DateFormat.EU:
            if self.force_aligned:
                return validators.DateEUAligned
            else:
                return validators.DateEU
        else:
            if self.force_aligned:
                return validators.DateAligned
            else:
                return validators.Date

    def extract(
        self,
        sheet: SheetReader,
        idx: int,
        validator: typing.Optional[typing.Type[validators.BaseValueModel]] = None,
        parser_row_offset: typing.Optional[int] = None,
        area_row_offset: typing.Optional[int] = None,
    ) -> typing.Any:
        if self.composed:
            year_date = self.composed.year.extract(
                sheet, idx, parser_row_offset=parser_row_offset, area_row_offset=area_row_offset
            )
            month_date = self.composed.month.extract(
                sheet, idx, parser_row_offset=parser_row_offset, area_row_offset=area_row_offset
            )
            return date(year_date.year, month_date.month, 1)
        else:
            return super().extract(sheet, idx, validator, parser_row_offset, area_row_offset)


rebuild_dataclass(ComposedDate, force=True)


@dataclass(config=PydanticConfig)
class ValueSource(JsonEncorder, ContentExtractorMixin):
    source: Source
    fallback: typing.Optional["ValueSource"] = None
    allow_negative: bool = False
    extract_params: ExtractParams = field(default_factory=lambda: ExtractParams())
    cleanup_during_header_processing: bool = True
    role: typing.Literal[Role.VALUE] = Role.VALUE

    @property
    def validator(self) -> typing.Optional[typing.Type[validators.BaseValueModel]]:
        if self.allow_negative:
            return validators.ValueNegative
        else:
            return validators.Value


@dataclass(config=PydanticConfig)
class VoidSource(JsonEncorder, ContentExtractorMixin):
    source: Source
    extract_params: ExtractParams = field(default_factory=lambda: ExtractParams())
    fallback: typing.Optional["VoidSource"] = None
    cleanup_during_header_processing: bool = True
    role: typing.Literal[Role.VOID] = Role.VOID


@dataclass(config=PydanticConfig)
class AuthorsSource(JsonEncorder, ContentExtractorMixin):
    source: Source
    extract_params: ExtractParams = field(default_factory=lambda: ExtractParams())
    fallback: typing.Optional["AuthorsSource"] = None
    cleanup_during_header_processing: bool = True
    role: typing.Literal[Role.AUTHORS] = Role.AUTHORS

    def get_validator(
        self, validator: typing.Optional[typing.Type[validators.BaseValueModel]]
    ) -> typing.Optional[typing.Type[validators.BaseValueModel]]:
        return validators.AuthorsValidator


@dataclass(config=PydanticConfig)
class PublicationDateSource(JsonEncorder, ContentExtractorMixin):
    source: Source
    extract_params: ExtractParams = field(default_factory=lambda: ExtractParams())
    fallback: typing.Optional["PublicationDateSource"] = None
    cleanup_during_header_processing: bool = True
    role: typing.Literal[Role.PUBLICATION_DATE] = Role.PUBLICATION_DATE



import typing
from abc import ABCMeta, abstractmethod

from nltk import stem
from pydantic import Field, ValidationError
from pydantic.dataclasses import dataclass, rebuild_dataclass
from typing_extensions import Annotated
from unidecode import unidecode

from . import validators
from .coordinates import Coord, CoordRange
from .errors import TableException
from .reader import SheetReader
from .utils import JsonEncorder, PydanticConfig

stemmer = stem.PorterStemmer()


class BaseCondition(metaclass=ABCMeta):
    @abstractmethod
    def check(
        self,
        sheet: SheetReader,
        parser_row_offset: typing.Optional[int] = None,
        area_row_offset: typing.Optional[int] = None,
    ) -> bool:
        pass


class ArithmeticsMixin:
    def __invert__(self):
        return NegCondition(self)

    def __or__(self, other):
        return OrCondition([self, other])

    def __and__(self, other):
        return AndCondition([self, other])


@dataclass(config=PydanticConfig)
class NegCondition(ArithmeticsMixin, BaseCondition, JsonEncorder):
    cond: "Condition"

    kind: typing.Literal["neg"] = "neg"

    def check(
        self,
        sheet: SheetReader,
        parser_row_offset: typing.Optional[int] = None,
        area_row_offset: typing.Optional[int] = None,
    ) -> bool:
        return not self.cond.check(sheet, parser_row_offset, area_row_offset)


@dataclass(config=PydanticConfig)
class AndCondition(ArithmeticsMixin, BaseCondition, JsonEncorder):
    conds: typing.List["Condition"]

    kind: typing.Literal["and"] = "and"

    def check(
        self,
        sheet: SheetReader,
        parser_row_offset: typing.Optional[int] = None,
        area_row_offset: typing.Optional[int] = None,
    ) -> bool:
        return all(e.check(sheet, parser_row_offset, area_row_offset) for e in self.conds)


@dataclass(config=PydanticConfig)
class OrCondition(ArithmeticsMixin, BaseCondition, JsonEncorder):
    conds: typing.List["Condition"]

    kind: typing.Literal["or"] = "or"

    def check(
        self,
        sheet: SheetReader,
        parser_row_offset: typing.Optional[int] = None,
        area_row_offset: typing.Optional[int] = None,
    ) -> bool:
        return any(e.check(sheet, parser_row_offset, area_row_offset) for e in self.conds)


@dataclass(config=PydanticConfig)
class RegexCondition(ArithmeticsMixin, BaseCondition, JsonEncorder):
    pattern: typing.Pattern
    coord: typing.Union[Coord, CoordRange]

    kind: typing.Literal["regex"] = "regex"

    def check(
        self,
        sheet: SheetReader,
        parser_row_offset: typing.Optional[int] = None,
        area_row_offset: typing.Optional[int] = None,
    ) -> bool:
        try:
            coord = self.coord

            if isinstance(coord, CoordRange):
                return any(
                    bool(
                        self.pattern.match(coord.content(sheet, parser_row_offset, area_row_offset))
                    )
                    for coord in coord
                )

            return bool(
                self.pattern.match(coord.content(sheet, parser_row_offset, area_row_offset))
            )
        except TableException:
            return False


@dataclass(config=PydanticConfig)
class IsDateCondition(ArithmeticsMixin, BaseCondition, JsonEncorder):
    coord: typing.Union[Coord, CoordRange]
    date_format: typing.Optional[str] = None

    kind: typing.Literal["is_date"] = "is_date"

    @property
    def validator_class(self) -> typing.Type[validators.BaseValueModel]:
        if not self.date_format:
            return validators.Date
        else:
            return validators.gen_date_format_validator(self.date_format)

    def check(
        self,
        sheet: SheetReader,
        parser_row_offset: typing.Optional[int] = None,
        area_row_offset: typing.Optional[int] = None,
    ) -> bool:
        coord = self.coord

        # Handle coord ranges
        if isinstance(coord, CoordRange):
            for c in coord:
                try:
                    content = c.content(sheet, parser_row_offset, area_row_offset)
                    content = content and content.strip()
                    self.validator_class(value=content)
                except ValidationError:
                    continue
                except (TableException, IndexError):
                    return False  # end was reached

                return True

        # Handle single coord
        try:
            content = coord.content(sheet, parser_row_offset, area_row_offset)
            content = content and content.strip()
            self.validator_class(value=content)
        except (TableException, ValidationError, IndexError):
            return False

        return True


@dataclass(config=PydanticConfig)
class StemmerCondition(ArithmeticsMixin, BaseCondition, JsonEncorder):
    """Compare content based on Porter Stemming algorithm"""

    content: str
    coord: typing.Union[Coord, CoordRange]

    kind: typing.Literal["stemmer"] = "stemmer"

    def _convert(self, text: str) -> str:
        return stemmer.stem(unidecode(text.strip()).lower())

    def __post_init__(self):
        self.content = self._convert(self.content)

    def check(
        self,
        sheet: SheetReader,
        parser_row_offset: typing.Optional[int] = None,
        area_row_offset: typing.Optional[int] = None,
    ) -> bool:
        try:
            if isinstance(self.coord, CoordRange):
                return any(
                    self._convert(coord.content(sheet, parser_row_offset, area_row_offset))
                    == self.content
                    for coord in self.coord
                )
            return (
                self._convert(self.coord.content(sheet, parser_row_offset, area_row_offset))
                == self.content
            )
        except TableException:
            return False


@dataclass(config=PydanticConfig)
class SheetNameRegexCondition(ArithmeticsMixin, BaseCondition, JsonEncorder):
    """Compare sheet name against the regex"""

    pattern: typing.Pattern

    kind: typing.Literal["sheet_name"] = "sheet_name"

    def check(
        self,
        sheet: SheetReader,
        parser_row_offset: typing.Optional[int] = None,
        area_row_offset: typing.Optional[int] = None,
    ) -> bool:
        if sheet.name is not None:
            return bool(self.pattern.match(sheet.name))
        else:
            return False


@dataclass(config=PydanticConfig)
class SheetIdxCondition(ArithmeticsMixin, BaseCondition, JsonEncorder):
    """Limit sheet by its index"""

    min: typing.Optional[int] = None
    max: typing.Optional[int] = None

    kind: typing.Literal["sheet_idx"] = "sheet_idx"

    def check(
        self,
        sheet: SheetReader,
        parser_row_offset: typing.Optional[int] = None,
        area_row_offset: typing.Optional[int] = None,
    ) -> bool:
        if self.min is not None and sheet.sheet_idx < self.min:
            return False

        if self.max is not None and sheet.sheet_idx > self.max:
            return False

        return True


@dataclass(config=PydanticConfig)
class SheetExtraCondition(ArithmeticsMixin, BaseCondition, JsonEncorder):
    """Checks whether sheet extra attribute matches"""

    field_name: str
    value: typing.Any

    kind: typing.Literal["sheet_extra"] = "sheet_extra"

    def check(
        self,
        sheet: SheetReader,
        parser_row_offset: typing.Optional[int] = None,
        area_row_offset: typing.Optional[int] = None,
    ) -> bool:
        if sheet.extra is None:
            return False
        if self.field_name in sheet.extra:
            return sheet.extra[self.field_name] == self.value
        else:
            return False


Condition = Annotated[
    typing.Union[
        NegCondition,
        OrCondition,
        AndCondition,
        RegexCondition,
        StemmerCondition,
        SheetNameRegexCondition,
        SheetIdxCondition,
        IsDateCondition,
        SheetExtraCondition,
    ],
    Field(discriminator="kind"),
]


# Need to update forward refs
rebuild_dataclass(NegCondition, force=True)
rebuild_dataclass(AndCondition, force=True)
rebuild_dataclass(OrCondition, force=True)
